@@ -150,8 +150,12 @@ class OpenSlide(AbstractSlide):
     def __init__(self, filename):
         """Open a whole-slide image."""
         AbstractSlide.__init__(self)
+        self._filename = filename
         self._osr = lowlevel.open(filename)
 
+    def __repr__(self):
+        return '%s(%r)' % (self.__class__.__name__, self._filename)
+
     @classmethod
     def detect_format(cls, filename):
         """Return a string describing the format vendor of the specified file.
@@ -224,7 +228,7 @@ class _OpenSlideMap(Mapping):
         self._osr = osr
 
     def __repr__(self):
-        return repr(dict(self))
+        return '<%s %r>' % (self.__class__.__name__, dict(self))
 
     def __len__(self):
         return len(self._keys())
@@ -266,23 +270,39 @@ class ImageSlide(AbstractSlide):
 
         file can be a filename or a PIL.Image."""
         AbstractSlide.__init__(self)
+        self._file_arg = file
         if isinstance(file, Image.Image):
+            self._close = False
             self._image = file
         else:
+            self._close = True
             self._image = Image.open(file)
 
+    def __repr__(self):
+        return '%s(%r)' % (self.__class__.__name__, self._file_arg)
+
     @classmethod
     def detect_format(cls, filename):
         """Return a string describing the format of the specified file.
 
         If the file format is not recognized, return None."""
         try:
-            return Image.open(filename).format
+            img = Image.open(filename)
+            format = img.format
+            if hasattr(img, 'close'):
+                # Pillow >= 2.5.0
+                img.close()
+            return format
         except IOError:
             return None
 
     def close(self):
         """Close the slide object."""
+        if self._close:
+            if hasattr(self._image, 'close'):
+                # Pillow >= 2.5.0
+                self._image.close()
+            self._close = False
         self._image = None
 
     @property
